import { useRef, useState, useEffect } from 'react';

var ERR_INIT_STORE = 'initStore should be a function';
var ERR_PAYLOAD = 'payload should be an object or a function';
var ERR_USE_OUT_STORE = 'useOutStore passed to get() is not initialized';
var EMPTY_OBJ = {};

var NOOP = function NOOP() {
  return undefined;
};

var __DEV__ = process.env.NODE_ENV !== 'production';

var notObj = function notObj(val) {
  return Object.prototype.toString.call(val) !== '[object Object]';
};

var map = new WeakMap();

function create(initStore) {
  if (__DEV__ && typeof initStore !== 'function') throw new Error(ERR_INIT_STORE);
  var listeners = [];
  var store = initStore({
    get: function get(useOutStore) {
      if (typeof useOutStore === 'undefined') return store;
      var outStore = map.get(useOutStore);
      if (__DEV__ && !outStore) throw new Error(ERR_USE_OUT_STORE);
      return outStore;
    },
    set: function set(payload) {
      if (typeof payload === 'function') {
        payload = payload(store);
      } else if (__DEV__ && notObj(payload)) {
        throw new Error(ERR_PAYLOAD);
      }

      Object.assign(store, payload);
      listeners.forEach(function (listener) {
        return listener(payload);
      });
    }
  });

  var useStore = function useStore() {
    var piece = useRef(EMPTY_OBJ);
    var onMount = useRef(NOOP);

    var _useState = useState(function () {
      var hasVal = false;
      var hasUpdate = false;
      var handler = {
        get: function get(target, key) {
          var val = store[key];

          if (typeof val !== 'function') {
            hasVal = true;
            target[key] = val;
            return target[key];
          }

          target[key] = new Proxy(val, {
            get: function get(fn, fnKey) {
              if (fnKey === 'loading' && !('loading' in fn)) fn.loading = false;
              return fn[fnKey];
            },
            apply: function apply(fn, _this, args) {
              var res = fn.apply(void 0, args);

              if (!('loading' in fn) || !res || typeof res.then !== 'function') {
                target[key] = fn;
                return res;
              }

              var setLoading = function setLoading(loading) {
                target[key].loading = loading;
                setState(function (s) {
                  return !s;
                });
              };

              target[key] = function () {
                var newRes = fn.apply(void 0, arguments);
                setLoading(true);
                return newRes.finally(function () {
                  return setLoading(false);
                });
              };

              setLoading(true);
              return res.finally(function () {
                return setLoading(false);
              });
            }
          });
          return target[key];
        },
        set: function set(target, key, val) {
          if (key in target && val !== target[key]) {
            hasUpdate = true;
            target[key] = val;
          }

          return true;
        }
      };
      piece.current = new Proxy({}, handler);

      var listener = function listener(partialState) {
        Object.assign(piece.current, partialState);

        if (hasUpdate) {
          hasUpdate = false;
          setState(function (s) {
            return !s;
          });
        }
      };

      listeners.push(listener);

      var unsubscribe = function unsubscribe() {
        listeners.splice(listeners.indexOf(listener), 1);
      };

      onMount.current = function () {
        handler.get = function (target, key) {
          return target[key];
        };

        return hasVal ? unsubscribe : unsubscribe();
      };

      return false;
    }),
        setState = _useState[1];

    useEffect(function () {
      return onMount.current();
    }, []);
    return piece.current;
  };

  map.set(useStore, store);
  return useStore;
}

export { create as default };
